@@ -162,7 +162,8 @@ void Flattener::printHelp(ostream& os) const {
      << "    Enable only the given comma-separated output sections." << std::endl
      << "  --not-sections <section_1,...section_n>" << std::endl
      << "    Disable the given comma-separated output sections." << std::endl
-     << "  -Werror\n    Turn warnings into errors" << std::endl;
+     << "  -Werror\n    Turn warnings into errors" << std::endl
+     << "  -w --disable-warnings\n    Supress all warnings" << std::endl;
 }
 
 bool Flattener::processOption(int& i, std::vector<std::string>& argv,
@@ -297,6 +298,8 @@ bool Flattener::processOption(int& i, std::vector<std::string>& argv,
     // Parsed by reference
   } else if (cop.getOption("-Werror")) {
     _flags.werror = true;
+  } else if (cop.getOption("-w --disable-warnings")) {
+    _fopts.supressWarnings = true;
   } else if (cop.getOption("--use-gecode")) {
 #ifdef HAS_GECODE
     _flags.twoPass = true;
@@ -627,9 +630,9 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
           std::string smm_compressed =
               FileUtils::encode_base64(FileUtils::deflate_string(smm_oss.str()));
           auto* ti = new TypeInst(Location().introduce(), Type::parstring(), nullptr);
+          auto* sl = new StringLit(Location().introduce(), smm_compressed);
           auto* checkString =
-              new VarDecl(Location().introduce(), ti, ASTString("_mzn_solution_checker"),
-                          new StringLit(Location().introduce(), smm_compressed));
+              new VarDecl(Location().introduce(), ti, ASTString("_mzn_solution_checker"), sl);
           auto* checkStringI = VarDeclI::a(Location().introduce(), checkString);
           env->output()->addItem(checkStringI);
 
@@ -646,9 +649,9 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
               std::string smm_stats_compressed =
                   FileUtils::encode_base64(FileUtils::deflate_string(smm_stats_oss.str()));
               auto* ti = new TypeInst(Location().introduce(), Type::parstring(), nullptr);
+              auto* sl = new StringLit(Location().introduce(), smm_stats_compressed);
               auto* checkStatsString =
-                  new VarDecl(Location().introduce(), ti, ASTString("_mzn_stats_checker"),
-                              new StringLit(Location().introduce(), smm_stats_compressed));
+                  new VarDecl(Location().introduce(), ti, ASTString("_mzn_stats_checker"), sl);
               auto* checkStatsStringI = VarDeclI::a(Location().introduce(), checkStatsString);
               env->output()->addItem(checkStatsStringI);
             }
