@@ -594,7 +594,8 @@ Expression* JSONParser::parseObject(istream& is, TypeInst* ti) {
       }
       return parseSet(is, ti);
     }
-    if (ti != nullptr && (ti->isEnum() || ti->type().bt() == Type::BT_UNKNOWN) &&
+    if (ti != nullptr &&
+        (ti->isEnum() || ti->type().bt() == Type::BT_INT || ti->type().bt() == Type::BT_UNKNOWN) &&
         (key == "e" || key == "i" || key == "c")) {
       if (!fields.empty()) {
         throw JSONError(_env, errLocation(), "invalid enum object");
@@ -605,8 +606,8 @@ Expression* JSONParser::parseObject(istream& is, TypeInst* ti) {
     auto it = fieldTIs.find(key);
     Expression* e = parseExp(is, true, it != fieldTIs.end() ? it->second : nullptr);
 
-    fields.push_back(new VarDecl(
-        Location().introduce(), new TypeInst(Location().introduce(), Expression::type(e)), key, e));
+    fields.push_back(
+        new VarDecl(Location().introduce(), new TypeInst(Location().introduce(), Type()), key, e));
     next = readToken(is);
   } while (next.t == T_COMMA);
   if (next.t != T_OBJ_CLOSE) {
@@ -759,7 +760,11 @@ Expression* JSONParser::coerceArray(TypeInst* ti, ArrayLit* al) {
             dims.emplace_back(1, nal->size());
           } else {
             if (nal->size() != dims[it.size() - 1].second) {
-              // TODO: Inconsistent array size.
+              throw JSONError(_env, Expression::loc(expr),
+                              "Mismatch in array dimensions in JSON array literal: dimension " +
+                                  std::to_string(it.size()) + " was found with both length " +
+                                  std::to_string(dims[it.size() - 1].second) + " and length " +
+                                  std::to_string(nal->size()) + ".");
             }
           }
         } else {
